import logging
from copy import deepcopy
from dataclasses import dataclass
from datetime import date

import pytest

import sdmx.model as model
from sdmx.model import common, v21
from sdmx.model.common import (
    Agency,
    AnnotableArtefact,
    BaseAnnotation,
    Component,
    ComponentList,
    Contact,
    IdentifiableArtefact,
    Item,
    ItemScheme,
    NameableArtefact,
    Representation,
)
from sdmx.testing import CompareTests


class TestBaseAnnotation:
    def test_text(self) -> None:
        """`text` can be :class:`str`."""
        a = BaseAnnotation(text="Foo")

        assert a.text.localizations["en"] == "Foo"


class TestAnnotableArtefact:
    @pytest.fixture
    def aa(self) -> AnnotableArtefact:
        return AnnotableArtefact(
            annotations=[
                v21.Annotation(id="foo", text="bar"),
                v21.Annotation(id="baz", title="baz_title", text="baz_text"),
            ]
        )

    def test_compare(self, aa: AnnotableArtefact) -> None:
        aa_other = AnnotableArtefact()

        assert False is aa.compare(aa_other)

    def test_get_annotation(self, aa: AnnotableArtefact) -> None:
        with pytest.raises(KeyError):
            aa.get_annotation(id="bar")

        # Retrieve with 1 key
        assert "bar" == str(aa.get_annotation(id="foo").text)

        # Retrieve with 2 keys
        assert "baz_text" == str(aa.get_annotation(id="baz", title="baz_title").text)

        # Annotations are not removed
        assert 2 == len(aa.annotations)

    def test_pop_annotation(self):
        aa = AnnotableArtefact()
        anno = v21.Annotation(id="foo", text="bar")

        assert 0 == len(aa.annotations)
        aa.annotations.append(anno)
        assert 1 == len(aa.annotations)

        with pytest.raises(KeyError):
            aa.pop_annotation(id="baz")

        assert anno == aa.pop_annotation(id="foo")
        assert 0 == len(aa.annotations)

    def test_eval_annotation(self, caplog) -> None:
        aa = AnnotableArtefact()

        value = dict(foo=1.1, bar=2, baz=True)

        aa.annotations.append(v21.Annotation(id="test-anno0", text=repr(value)))
        aa.annotations.append(v21.Annotation(id="test-anno1", text="value['foo']"))

        # Returns None for a missing ID
        assert None is aa.eval_annotation(id="wrong-id")

        # Entire value is retrieved
        assert value == aa.eval_annotation(id="test-anno0")

        # Cannot handle a variable reference with no globals;
        assert "value['foo']" == aa.eval_annotation(id="test-anno1")
        assert "name 'value' is not defined" in caplog.messages[0]
        caplog.clear()

        # Globals can be used if passed
        assert value["foo"] == aa.eval_annotation(
            id="test-anno1", globals=dict(value=value)
        )


URN = "urn:sdmx:org.sdmx.infomodel.conceptscheme.ConceptScheme=IT1:VARIAB_ALL(9.6)"


class TestIdentifiableArtefact(CompareTests):
    @pytest.fixture
    def obj(self) -> common.IdentifiableArtefact:
        return common.IdentifiableArtefact(
            id="VARIAB_ALL",
            uri="https://example.com",
            urn=URN,
        )

    @pytest.mark.parametrize(
        "callback",
        (
            None,
            lambda obj: setattr(obj, "id", "VARIAB_ALL_"),
            lambda obj: setattr(obj, "uri", "https://example.com/foo"),
            lambda obj: setattr(obj, "urn", URN.replace("VARIAB_ALL", "X")),
        ),
    )
    def test_compare(self, obj: common.IdentifiableArtefact, callback) -> None:
        super().test_compare(obj, callback)

    def test_init_empty_id(self):
        """IdentifiableArtefact can be initialized with an empty :class:`str` as ID."""
        # No id= parameter → id attribute is MissingID
        ia0 = IdentifiableArtefact()
        assert common.MissingID == ia0.id
        assert common.MissingID is ia0.id

        # Empty string parameter → id attribute is empty string
        ia1 = IdentifiableArtefact(id="")
        assert "" == ia1.id

    def test_init_urn(self):
        """IdentifiableArtefact can be initialized with URN."""
        ia = IdentifiableArtefact(urn=URN)
        assert "VARIAB_ALL" == ia.id

    def test_general(self) -> None:
        urn_pat = URN.replace("(", r"\(").replace(")", r"\)")

        with pytest.raises(
            ValueError, match=f"ID BAD_URN does not match URN {urn_pat}"
        ):
            IdentifiableArtefact(id="BAD_URN", urn=URN)

        # IdentifiableArtefact is hashable
        ia = IdentifiableArtefact()
        assert hash(ia) == id(ia)

        ia = IdentifiableArtefact(id="foo")
        assert hash(ia) == hash("foo")

        # Subclass is hashable
        ad = model.AttributeDescriptor()
        assert hash(ad) == id(ad)

    def test_gt_lt(self):
        """Test IdentifiableArtefact.__gt__ and IdentifiableArtefact.__lt__."""
        # IdentifiableArtefact of the same class can be sorted
        items = [Item(id="b"), Item(id="a")]
        assert list(reversed(items)) == sorted(items)

        # Mixed IdentifiableArtefact and str can be sorted
        assert [Item(id="a"), "b", "c"] == sorted(["b", Item(id="a"), "c"])

    def test_hash_subclass(self):
        @dataclass
        class Foo(IdentifiableArtefact):
            __hash__ = IdentifiableArtefact.__hash__

        f = Foo(id="FOO")
        assert hash("FOO") == hash(f)


class TestNameableArtefact:
    def test_compare(self, caplog) -> None:
        na1 = NameableArtefact(name=dict(en="Name"), description=dict(en="Description"))
        na2 = NameableArtefact()

        assert not na1.compare(na2)
        assert "Not identical: NameableArtefact.name=en: Name != " in caplog.messages
        caplog.clear()

        na2.name["en"] = "Name"

        assert not na1.compare(na2)
        assert (
            "Not identical: NameableArtefact.description=en: Description != "
            in caplog.messages
        )

        na2.description["en"] = "Description"

        assert na1.compare(na2)

    def test_eq(self):
        na = NameableArtefact(id="FOO", name="Foo")
        assert na == "FOO"


class TestVersionableArtefact(CompareTests):
    @pytest.fixture
    def obj(self) -> common.VersionableArtefact:
        return common.VersionableArtefact(
            valid_from="0001-01-01",
            valid_to="9999-12-31",
            version="1.0.0",
        )

    @pytest.mark.parametrize(
        "callback",
        (
            None,
            lambda obj: setattr(obj, "valid_from", "0001-01-02"),
            lambda obj: setattr(obj, "valid_to", "9999-12-30"),
            lambda obj: setattr(obj, "version", "1.0.1"),
        ),
    )
    def test_compare(self, obj: common.VersionableArtefact, callback) -> None:
        super().test_compare(obj, callback)

    def test_urn(self) -> None:
        va = common.VersionableArtefact(id="VARIAB_ALL", urn=URN)

        # Version is parsed from URN
        assert va.version == "9.6"

        # Mismatch raises an exception
        with pytest.raises(ValueError, match="Version '9.7' does not match URN"):
            common.VersionableArtefact(version="9.7", urn=URN)

    def test_version_none(self) -> None:
        va = common.VersionableArtefact(version="None")
        assert va.version is None


class TestMaintainableArtefact(CompareTests):
    @pytest.fixture
    def obj(self) -> common.MaintainableArtefact:
        return common.MaintainableArtefact(
            is_external_reference=True,
            is_final=True,
            maintainer=common.Agency(id="A1"),
            service_url="https://example.com",
            structure_url="https://example.com",
        )

    @pytest.mark.parametrize(
        "callback",
        (
            None,
            lambda obj: setattr(obj, "is_external_reference", False),
            lambda obj: setattr(obj, "is_final", False),
            lambda obj: setattr(obj, "maintainer", common.Agency(id="A2")),
            lambda obj: setattr(obj, "service_url", "https://example.com/foo"),
            lambda obj: setattr(obj, "structure_url", "https://example.com/foo"),
        ),
    )
    def test_compare(self, obj: common.MaintainableArtefact, callback) -> None:
        super().test_compare(obj, callback)

    def test_urn(self) -> None:
        ma = common.MaintainableArtefact(id="VARIAB_ALL", urn=URN)

        # Maintainer is parsed from URN
        assert ma.maintainer == Agency(id="IT1")

        # Mismatch raises an exception
        with pytest.raises(ValueError, match="Maintainer FOO does not match URN"):
            common.MaintainableArtefact(maintainer=Agency(id="FOO"), urn=URN)


class TestItem:
    def test_name(self) -> None:
        """`name` can be :class:`str`."""
        Item(name="Foo")

    def test_general(self):
        # Add a tree of 10 items
        items = []
        for i in range(10):
            items.append(Item(id="Foo {}".format(i)))

            if i > 0:
                items[-1].parent = items[-2]
                items[-2].child.append(items[-1])

        # __init__(parent=...)
        Item(id="Bar 1", parent=items[0])
        assert len(items[0].child) == 2

        # __init__(child=)
        bar2 = Item(id="Bar 2", child=[items[0]])

        # __contains__()
        assert items[0] in bar2
        assert items[-1] in items[0]

        # get_child()
        assert items[0].get_child("Foo 1") == items[1]

        with pytest.raises(ValueError):
            items[0].get_child("Foo 2")

        # Hierarchical IDs constructed automatically
        assert items[0].child[0].hierarchical_id == "Bar 2.Foo 0.Foo 1"


class TestItemScheme:
    def test_compare0(self, caplog):
        caplog.set_level(logging.DEBUG)

        is0 = ItemScheme()
        is1 = ItemScheme()

        is0.append(Item(id="foo", name="Foo"))
        is1.append(Item(id="foo", name="Bar"))

        assert not is0.compare(is1)

        # Log shows that items with same ID have different name
        assert "Not identical: Item.name=en: Foo != en: Bar" in caplog.messages
        assert (
            "Not identical: ItemScheme.items={'foo': <Item foo: Foo>} != {'foo': <Item foo: Bar>}"
            in caplog.messages
        )

    def test_compare1(self) -> None:
        is0: ItemScheme = ItemScheme(id="is0")
        is0.append(Item(id="foo"))

        is1: ItemScheme = ItemScheme(id="is0")
        is1.append(Item(id="foo"))

        assert is0.compare(is1) is True

        is2: ItemScheme = ItemScheme(id="is0")
        is2.append(Item(id="bar"))

        assert is0.compare(is2) is False

    def test_deepcopy(self) -> None:
        """ItemScheme may be :func:`.deepcopy`'d."""
        is0: ItemScheme = ItemScheme(id="is0")

        deepcopy(is0)

    def test_get_hierarchical(self) -> None:
        is0: ItemScheme = ItemScheme(id="is0")
        foo: Item = Item(id="foo")
        bar: Item = Item(id="bar")
        foo.append_child(bar)

        is0.append(foo)
        is0.append(bar)

        assert bar is is0.get_hierarchical("foo.bar")

        # With non-hierarchical ID
        assert foo is is0.get_hierarchical("foo")

        with pytest.raises(KeyError):
            is0.get_hierarchical("foo.baz")

    def test_other(self) -> None:
        is0: ItemScheme = ItemScheme(id="is0")
        foo0: Item = Item(id="foo0")

        # With a single Item

        # append()
        is0.append(foo0)

        # __getattr__
        assert is0.foo0 is foo0

        # __getitem__
        assert is0["foo0"] is foo0

        # __contains__
        assert "foo0" in is0
        assert foo0 in is0

        # __len__
        assert len(is0) == 1

        # __repr__
        assert repr(is0) == "<ItemScheme is0 (1 items)>"

        # __iter__
        assert all(i is foo0 for i in is0)

        # With multiple Items

        foo1: Item = Item(id="foo1")
        foo2: Item = Item(id="foo2")
        items_list = [foo0, foo1, foo2]
        items_dict = {"foo0": foo0, "foo1": foo1, "foo2": foo2}

        # set with a non-dict
        is0.items = {i.id: i for i in items_list}
        assert is0.items == items_dict

        # set with a dict
        is0.items = items_dict
        assert is0.items == items_dict

        # extend()
        is0.items = {foo0.id: foo0}
        is0.extend(items_list[1:])
        assert is0.items == items_dict

        # setdefault()
        bar0 = is0.setdefault(id="bar")
        assert bar0.id == "bar"

        with pytest.raises(ValueError):
            is0.setdefault(foo0, id="bar")

        is0.setdefault(id="bar1", parent="foo0")
        bar1 = is0.setdefault(id="bar1", parent=foo0)

        # get_hierarchical()
        assert is0.get_hierarchical("foo0.bar1") is bar1


class TestRepresentation:
    def test_repr(self) -> None:
        is0: ItemScheme = ItemScheme(id="is0")
        r = Representation(enumerated=is0)
        assert "<Representation: is0, []>" == repr(r)


class TestComponentList:
    def test_compare(self, caplog) -> None:
        """Test comparison of two CL with mismatched components."""

        components = [Component(id=s) for s in ("FOO", "BAR", "BAZ")]

        cl1: ComponentList = ComponentList(id="CL", components=components)
        cl2: ComponentList = ComponentList(id="CL", components=components[:-1])

        # cl1 and cl2 compare as different
        assert False is cl1.compare(cl2)

        # commented: This message is currently disabled
        # # Log message is emitted for mismatched components
        # assert "CL has no component with ID 'BAZ'" in caplog.messages


class TestContact:
    def test_init(self):
        Contact(
            name="Jane Smith", org_unit="Statistics Office", responsibility="Director"
        )

        c1 = Contact()
        c2 = Contact()

        c2.name = "foo"

        assert repr(c1) != repr(c2)


class TestStructure(CompareTests):
    @pytest.fixture
    def obj(self):
        # Use BaseDataStructureDefinition as a concrete/less abstract subclass
        return common.BaseDataStructureDefinition()

    @pytest.mark.parametrize(
        "callback",
        (
            None,
            lambda obj: obj.attributes.getdefault(id="a1"),
            lambda obj: obj.dimensions.getdefault(id="d1"),
        ),
    )
    def test_compare(self, obj: common.Structure, callback) -> None:
        """:py:`compare(…)` is :any:`False` when attributes are changed."""
        super().test_compare(obj, callback)

    def test_grouping(self, obj) -> None:
        # Class has an AttributeDescriptor, DimensionDescriptor
        assert 2 == len(obj.grouping)

        # With GroupDimensionDescriptor, these are each included separately
        obj.group_dimensions = {
            "FOO": model.GroupDimensionDescriptor(id="FOO"),
            "BAR": model.GroupDimensionDescriptor(id="BAR"),
        }
        assert 4 == len(obj.grouping)

    def test_replace_grouping(self, obj) -> None:
        class Foo(model.ComponentList):
            pass

        # Cannot replace with an instance of Foo, because this does not correspond to
        # the type of any member of the class
        with pytest.raises(TypeError, match="No grouping of type"):
            obj.replace_grouping(Foo())


class TestAttributeValue(CompareTests):
    @pytest.fixture
    def obj(self) -> common.AttributeValue:
        return common.AttributeValue(
            start_date=date.fromisoformat("0001-01-01"),
            value="v1",
            value_for=common.DataAttribute(id="da1"),
        )

    @pytest.mark.parametrize(
        "callback",
        (
            None,
            lambda obj: setattr(obj, "start_date", date.fromisoformat("0001-01-02")),
            lambda obj: setattr(obj, "value", "v2"),
            lambda obj: setattr(obj, "value_for", common.DataAttribute(id="da2")),
        ),
    )
    def test_compare(self, obj: common.AttributeValue, callback) -> None:
        """:py:`compare(…)` is :any:`False` when attributes are changed."""
        super().test_compare(obj, callback)


class TestBaseObservation(CompareTests):
    @pytest.fixture
    def obj(self) -> common.BaseObservation:
        return common.BaseObservation(
            attached_attribute={"FOO": common.AttributeValue(value="f1")},
            dimension=common.Key(BAR="b1"),
            group_keys={common.GroupKey(id="g1")},
            series_key=common.SeriesKey(),
            value=1.0,
        )

    @pytest.mark.parametrize(
        "callback",
        (
            None,
            lambda o: o.attached_attribute.popitem(),
            lambda o: setattr(o, "dimension", common.Key(BAR="b2")),
            lambda o: setattr(o, "group_keys", {common.GroupKey(id="g2")}),
            lambda o: setattr(o, "series_key", None),
            lambda o: setattr(o, "value", 1.1),
        ),
    )
    def test_compare(self, obj: common.BaseObservation, callback) -> None:
        """:py:`compare(…)` is :any:`False` when attributes are changed.

        NB for .value_for, see test_v{21,30}.TestObservation.test_compare.
        """
        super().test_compare(obj, callback)


class TestBaseDataSet(CompareTests):
    @pytest.fixture
    def obj(self) -> common.BaseDataSet:
        k = common.Key(FOO=common.KeyValue(id="FOO", value="f1"))
        o = common.BaseObservation(dimension=k, value=1.0)
        result = common.BaseDataSet(
            action=common.ActionType.information,
            valid_from="0001-01-01",
            obs=[o],
            group={common.GroupKey(): [o]},
            series={common.SeriesKey(): [o]},
        )
        return result

    @pytest.mark.parametrize(
        "callback",
        (
            None,
            # Change attributes
            lambda ds: setattr(ds, "action", common.ActionType.delete),
            lambda ds: ds.group.popitem(),  # Remove a group key
            lambda ds: ds.obs.pop(-1),  # Change *number* of obs
            lambda ds: setattr(ds.obs[-1], "value", 1.1),  # Change *value* of one obs
            lambda ds: ds.series.popitem(),  # Remove a series key
            lambda ds: setattr(ds, "valid_from", "0001-01-02"),
        ),
    )
    def test_compare(self, obj: common.BaseDataSet, callback) -> None:
        """:py:`compare(…)` is :any:`False` when attributes are changed."""
        super().test_compare(obj, callback)



"""SDMX-REST API v2.1.0.

Note that version 2.1.0 of the REST API corresponds to version 3.0.0 of the overall
SDMX standards. See the `documentation
<https://github.com/sdmx-twg/sdmx-rest/tree/v2.1.0/doc>`__ and `OpenAPI specification
<https://github.com/sdmx-twg/sdmx-rest/blob/v2.1.0/api/sdmx-rest.yaml>`__ for further
details.
"""

from collections import ChainMap

from . import common
from .common import PathParameter, QueryParameter, Resource

#: v2.1.0-specific path and query parameters.
PARAM: dict[str, common.Parameter] = {
    # Path parameters
    "component_id": PathParameter("component_id"),
    "context": PathParameter(
        "context", common.NAMES["context"] | {"metadataprovisionagreement"}
    ),
    "context_d": PathParameter(
        "context", {"datastructure", "dataflow", "provisionagreement", "*"}, "*"
    ),
    "provider_id": PathParameter("provider_id"),
    "version": PathParameter("version", set(), "+"),
    #
    # Query parameters
    "attributes": QueryParameter("attributes"),
    "c": QueryParameter("c"),
    "detail_m": QueryParameter("detail", {"allstubs", "full"}),
    "detail_s": QueryParameter("detail", common.NAMES["detail_s"] | {"raw"}),
    "measures": QueryParameter("measures"),
    "references_s": QueryParameter(
        "references", common.NAMES["references_s"] | {"ancestors"} | set(Resource)
    ),
    "updated_after": QueryParameter("update_after"),
    "updated_before": QueryParameter("update_before"),
}


class URL(common.URL):
    """Utility class to build SDMX-REST API v2.1.0 URLs."""

    _all_parameters = ChainMap(common.PARAM, PARAM)

    def handle_availability(self):
        """Handle URL parameters for availability endpoints."""
        self._params.setdefault("agency_id", self.source.id)
        self.handle_path_params(
            "availability/{context_d}/{agency_id}/{resource_id}/{version}/{key}/"
            "{component_id}"
        )
        self.handle_query_params("c mode references_a updated_after")

    def handle_data(self):
        """Handle URL parameters for data endpoints."""
        self._params.setdefault("agency_id", self.source.id)
        self.handle_path_params(
            self.rt + "/{context_d}/{agency_id}/{resource_id}/{version}/{key}"
        )
        self.handle_query_params(
            "c updated_after first_n_observations last_n_observations "
            "dimension_at_observation attributes measures include_history"
        )

    def handle_metadata(self):
        """Handle URL parameters for metadata endpoints."""
        self._path.update({"metadata": None})
        if self.resource_type == common.Resource.metadataflow:
            self.handle_path_params(
                "metadataflow/{agency_id}/{resource_id}/{version}/{provider_id}"
            )
        elif self.resource_type == common.Resource.metadata:
            self.handle_path_params("metadataset/{provider_id}/{resource_id}/{version}")
        else:
            self.handle_path_params(self.rt + "/{agency_id}/{resource_id}/{version}")
        self.handle_query_params("detail_s")

    def handle_registration(self):
        """Handle URL parameters for registration endpoints."""
        self.handle_path_params("registration")
        if "context" in self._params:
            self.handle_path_params("{context_d}/{agency_id}/{resource_id}/{version}")
            self.handle_query_params("updated_after updated_before")
        elif "agency_id" in self._params:
            self.handle_path_params("provider/{agency_id}/{provider_id}")
            self.handle_query_params("updated_after updated_before")
        else:
            self.handle_path_params("id/{resource_id}")  # "registrationID" in the spec
            # No query parameters

    def handle_structure(self):
        """Handle URL parameters for structure endpoints."""
        self.handle_path_params(f"structure/{self.rt}")
        super().handle_structure()



from dataclasses import dataclass
from typing import Optional

import pytest
from lxml.etree import Element

from sdmx.compare import Options, compare


class Foo:
    """Dummy class for testing."""

    pass


def test_allow_implied_urn() -> None:
    """Test :py:`except ValueError` block in compare function for dataclasses."""

    @dataclass
    class Bar:
        urn: Optional[str] = None

    b1 = Bar(urn="")
    b2 = Bar()

    compare(b1, b2, Options(base=b1))
    compare(b2, b1, Options(base=b2))


def test_not_implemented() -> None:
    f1 = Foo()
    f2 = Foo()

    with pytest.raises(NotImplementedError):
        compare(f1, f2, Options(base=f1))


def test_element() -> None:
    """Test :py:`except TypeError` block in compare function for lxml.etree.Element."""
    e = Element("foo")
    assert False is compare(e, Foo(), Options(base=e))



"""Test code appearing in the documentation.

The 'network' mark is used to mark tests that will access the Internet. In order to run
these tests, a command-line argument must be given:

$ pytest -m network [...]
"""

import logging
import re
from typing import TYPE_CHECKING

import numpy as np
import pandas as pd
import pytest

import sdmx
from sdmx import Client
from sdmx.dictlike import DictLike
from sdmx.model.v21 import GenericDataSet
from sdmx.testing import assert_pd_equal

log = logging.getLogger(__name__)

if TYPE_CHECKING:
    import sdmx.message


@pytest.mark.network
def test_example() -> None:
    """Code from :file:`doc/example.rst`."""
    import sdmx

    estat = sdmx.Client("ESTAT")

    sm: "sdmx.message.StructureMessage" = estat.datastructure("UNE_RT_A")

    # Identify partial URNs for some code lists
    partial_urns = []
    for cl in sm.codelist.values():
        if cl.id in ("AGE", "UNIT", "SEX"):
            partial_urns.append(cl.urn.rpartition("=")[2])
    # These strings should be the ones hard-coded in example.rst
    log.info(f"for cl in {repr(partial_urns).strip('[]')}:")

    # NB Use partial URNs to match even if only single versions are stored under keys
    #    like "AGE"
    for cl in partial_urns:
        print(sdmx.to_pandas(sm.get(cl)))

    dm = estat.data("UNE_RT_A", key={"geo": "EL+ES+IE"}, params={"startPeriod": "2007"})

    data = sdmx.to_pandas(dm).xs("Y15-74", level="age", drop_level=False)

    data.loc[("A", "Y15-74", "PC_ACT", "T")]

    # Further checks per https://github.com/dr-leo/pandaSDMX/issues/157

    # DimensionDescriptor for the structure message
    dd1 = sm.structure.UNE_RT_A.dimensions  # type: ignore [attr-defined]

    # DimensionDescriptor retrieved whilst validating the data message
    dd2 = dm.data[0].structured_by.dimensions

    # DimensionDescriptors have same ID, components and order
    assert dd1 == dd2

    # One SeriesKey from the data message
    sk = list(dm.data[0].series.keys())[0]

    # Key values have same order as in the DSD
    assert dd1.order_key(sk) == sk


@pytest.mark.network
def test_doc_index1() -> None:
    """A code example that formerly appeared in doc/index.rst."""
    estat = Client("ESTAT")
    sm0 = estat.dataflow("UNE_RT_A")

    with pytest.raises(TypeError):
        # This presumes the DataStructureDefinition instance can conduct a
        # network request for its own content
        sm0.dataflow.UNE_RT_A.structure(request=True, target_only=False)

    # Same effect
    sm1 = estat.get(
        "datastructure",
        sm0.dataflow.UNE_RT_A.structure.id,
        params=dict(references="descendants"),
    )
    assert isinstance(sm1, sdmx.message.StructureMessage)

    # Even better: Client.get(…) should examine the class and ID of the object
    # structure = estat.get(flow_response.dataflow.UNE_RT_A.structure)

    # Show some codelists
    s = sdmx.to_pandas(sm1)
    expected = pd.Series(
        {
            "ACP": "African, Caribbean and Pacific Group of States, signatories of the "
            "Partnership Agreement",
            "ACP_AFR": "African ACP states",
            "ACP_CRB": "Caribbean ACP states",
            "ACP_PAC": "Pacific ACP states",
            "AD": "Andorra",
        },
        name="Geopolitical entity (reporting)",
    ).rename_axis("GEO")

    # StructureMessage is converted to DictLike
    assert isinstance(s, DictLike)
    # "codelist" key retrieves a second-level DictLike
    assert isinstance(s.codelist, DictLike)  # type: ignore [attr-defined]

    # Same effect
    # NB At some times (e.g. between 2024-03-15 and 2024-06-18) this query retrieves
    #    multiple versions of similar artefacts. A more explicit argument to get() that
    #    includes the version (like get("GEO(21.0)")) may be temporarily needed.
    s = sdmx.to_pandas(sm1.get("ESTAT:GEO"))
    assert_pd_equal(s.sort_index().head(), expected)


@pytest.mark.network
def test_doc_usage_structure():
    """Code examples in walkthrough.rst."""
    ecb = Client("ECB")

    ecb_via_proxy = Client("ECB", proxies={"http": "http://1.2.3.4:5678"})
    assert all(
        getattr(ecb_via_proxy.session, k) == v
        for k, v in (
            ("proxies", {"http": "http://1.2.3.4:5678"}),
            ("stream", False),
            ("timeout", 30.0),
        )
    )

    msg1 = ecb.categoryscheme(agency_id="all")

    assert msg1.response.url == (
        "https://data-api.ecb.europa.eu/service/categoryscheme/all/all/latest"
        "?references=parentsandsiblings"
    )

    # Check specific headers
    headers = msg1.response.headers
    assert re.fullmatch(
        r"application/vnd\.sdmx\.structure\+xml; ?version=2\.1",
        headers["Content-Type"],
    )
    assert {"Connection", "Date", "Server"} <= set(headers)

    # Removed: in pandaSDMX 0.x this was a convenience method that (for this
    # structure message) returned two DataStructureDefinitions. Contra the
    # spec, that assumes:
    # - There is 1 Categorization using the CategoryScheme; there could be
    #   many.
    # - The Categorization maps DataStructureDefinitions to Categories, when
    #   there could be many.
    # list(cat_response.category_scheme['MOBILE_NAVI']['07'])

    dfs = sdmx.to_pandas(msg1.dataflow).head()
    expected = pd.Series(
        {
            "AME": "AMECO",
            "BKN": "Banknotes statistics",
            "BLS": "Bank Lending Survey Statistics",
            "BOP": (
                "Euro Area Balance of Payments and International Investment "
                "Position Statistics"
            ),
            "BSI": "Balance Sheet Items",
        }
    )
    assert_pd_equal(dfs, expected)

    flows = ecb.dataflow()  # noqa: F841
    dsd_id = msg1.dataflow.EXR.structure.id
    assert dsd_id == "ECB_EXR1"

    refs = dict(references="all")
    msg2 = ecb.datastructure(resource_id=dsd_id, params=refs)
    dsd = msg2.structure[dsd_id]

    assert sdmx.to_pandas(dsd.dimensions) == [
        "FREQ",
        "CURRENCY",
        "CURRENCY_DENOM",
        "EXR_TYPE",
        "EXR_SUFFIX",
        "TIME_PERIOD",
    ]

    cl = sdmx.to_pandas(msg2.codelist["CL_CURRENCY"]).sort_index()
    expected = pd.Series(
        {
            "ADF": "Andorran Franc (1-1 peg to the French franc)",
            "ADP": "Andorran Peseta (1-1 peg to the Spanish peseta)",
            "AED": "United Arab Emirates dirham",
            "AFA": "Afghanistan afghani (old)",
            "AFN": "Afghanistan, Afghanis",
        },
        name="Currency code list",
    ).rename_axis("CL_CURRENCY")
    assert_pd_equal(cl.head(), expected)


# @pytest.mark.skip(reason="Temporarily offline on 2021-03-23")
@pytest.mark.network
def test_doc_usage_data():
    """Code examples in usage.rst."""
    ecb = Client("ECB")

    data_response = ecb.data(
        resource_id="EXR",
        key={"CURRENCY": "USD+JPY"},
        params={"startPeriod": "2016", "endPeriod": "2016-12-31"},
    )
    # # Commented: do the same without triggering requests for validation
    # data_response = ecb.data(resource_id='EXR', key='.JPY+USD...',
    #                          params={'startPeriod': '2016',
    #                                  'endPeriod': '2016-12-31'})
    data = data_response.data[0]

    assert type(data) is GenericDataSet

    # This message doesn't explicitly specify the remaining dimensions; unless
    # they are inferred from the SeriesKeys, then the DimensionDescriptor is
    # not complete
    # assert data.structured_by.dimensions[-1] == 'TIME_PERIOD'
    # data.dim_at_obs

    series_keys = list(data.series)

    assert len(series_keys) == 22

    series_keys[5]

    assert sorted(set(sk.FREQ.value for sk in data.series)) == "A D H M Q".split()

    daily = sdmx.to_pandas(data).xs("D", level="FREQ")
    assert len(daily) == 514

    assert_pd_equal(
        daily.tail().values, np.array([1.0446, 1.0445, 1.0401, 1.0453, 1.0541])
    )


def test_doc_howto_timeseries(specimen):
    with specimen("sg-ts.xml") as f:
        ds = sdmx.read_sdmx(f).data[0]

    # Convert to pd.Series and unstack the time dimension to columns
    base = sdmx.to_pandas(ds)
    s1 = base.unstack("TIME_PERIOD")

    # DatetimeIndex on columns
    s1.columns = pd.to_datetime(s1.columns)
    assert isinstance(s1.columns, pd.DatetimeIndex)

    # DatetimeIndex on index
    s2 = base.unstack("TIME_PERIOD").transpose()
    s2.index = pd.to_datetime(s2.index)
    assert isinstance(s2.index, pd.DatetimeIndex)

    # Same with pd.PeriodIndex
    s3 = s1.to_period(axis=1)
    assert isinstance(s3.columns, pd.PeriodIndex)
    assert s3.columns.freqstr == "M"

    s4 = s2.to_period(axis=0)
    assert isinstance(s4.index, pd.PeriodIndex)
    assert s4.index.freqstr == "M"



from typing import TypedDict

from sdmx.convert.pandas import Attributes
from sdmx.format.csv.common import CSVFormatOptions
from sdmx.model.common import Agency


class VersionableArtefactArgs(TypedDict, total=False):
    version: str


class MaintainableArtefactArgs(VersionableArtefactArgs):
    maintainer: Agency


class ToCSVArgs(TypedDict, total=False):
    attributes: Attributes
    format_options: CSVFormatOptions



import re
from functools import partial
from itertools import product
from typing import TYPE_CHECKING, cast

import pandas as pd
import pytest

import sdmx
from sdmx import message
from sdmx.convert.pandas import Attributes
from sdmx.format import csv
from sdmx.format.csv.common import CSVFormatOptions, Labels
from sdmx.format.csv.v1 import FormatOptions as V1FormatOptions
from sdmx.format.csv.v2 import FormatOptions as V2FormatOptions
from sdmx.format.csv.v2 import Keys
from sdmx.model import common, v21

if TYPE_CHECKING:
    from pathlib import Path

    from sdmx.types import MaintainableArtefactArgs, ToCSVArgs


MARKS = {
    "ESTAT/esms.xml": pytest.mark.xfail(raises=NotImplementedError),
}


def _add_test_dsd(ds: v21.DataSet) -> None:
    if ds.described_by is None:
        dsd = ds.structured_by
        if dsd is None:
            pytest.skip(reason="No DFD or DSD")
        else:
            # Construct a fake/temporary DFD
            ds.described_by = v21.DataflowDefinition(
                id=f"_TEST_{dsd.id}", maintainer=dsd.maintainer, version="0.0"
            )


@pytest.mark.parametrize_specimens("path", kind="data", marks=MARKS)
@pytest.mark.parametrize("format_options", [None, V1FormatOptions(), V2FormatOptions()])
def test_write_data(tmp_path: "Path", specimen, path, format_options) -> None:
    if ("v3", "csv") == path.parts[-3:-1]:
        pytest.skip("SDMX-CSV 3.0.0 examples cannot be read without DSD")

    msg = cast("message.DataMessage", sdmx.read_sdmx(path))

    kw: "ToCSVArgs" = dict(attributes=Attributes.all, format_options=format_options)
    for i, dataset in enumerate(msg.data):
        _add_test_dsd(dataset)

        # Writer runs successfully
        result = sdmx.to_csv(dataset, rtype=pd.DataFrame, **kw)
        assert isinstance(result, (pd.Series, pd.DataFrame))

        # print(result.head().to_string()) # DEBUG

        # Write directly to file also works
        path_out = tmp_path.joinpath(f"{i}.csv")
        assert None is sdmx.to_csv(dataset, path=path_out, **kw)
        assert path_out.exists()

        # Standard features are respected
        if isinstance(format_options, V1FormatOptions):
            assert "DATAFLOW" == result.columns[0]
            assert "OBS_VALUE" in result.columns
            with open(path_out, "r") as f:
                assert f.readline().startswith("DATAFLOW,")


@pytest.fixture
def messages() -> tuple[message.StructureMessage, message.DataMessage]:
    """:class:`.DataMessage` with complete structures."""
    sm = message.StructureMessage()
    ma_args: "MaintainableArtefactArgs" = dict(
        maintainer=common.Agency(id="AGENCY"), version="1.0"
    )
    dsd = v21.DataStructureDefinition(id="DSD", **ma_args)
    df = v21.DataflowDefinition(id="DF", **ma_args, structure=dsd)
    dd, md, ad = dsd.dimensions, dsd.measures, dsd.attributes
    cs = common.ConceptScheme(id="CS", **ma_args)

    # Code lists
    cl: dict[str, common.Codelist] = dict()

    for concept_id in ("FOO", "BAR", "BAZ", "QUX"):
        cs.setdefault(id=concept_id, name=concept_id.title())
        _cl = cl[concept_id] = common.Codelist(id=f"CL_{concept_id}")
        [
            _cl.setdefault(id=f"{concept_id}{i}", name=f"{concept_id.title()} {i}")
            for i in range(3)
        ]
    # FOO and BAR are dimensions
    dd.getdefault(id="DIM_FOO", concept_identity=cs["FOO"])
    dd.getdefault(id="DIM_BAR", concept_identity=cs["BAR"])
    # The primary measure has an ID different from "OBS_VALUE"
    OBS_VALUE = cs.setdefault(id="OBS_VALUE_X", name="Observation value")
    md.getdefault(id="OBS_VALUE_X", concept_identity=OBS_VALUE)
    # BAZ and QUX are attributes
    ad.getdefault(id="ATTR_BAZ", concept_identity=cs["BAZ"])
    ad.getdefault(id="ATTR_QUX", concept_identity=cs["QUX"])

    ds = v21.DataSet(described_by=df, structured_by=dsd)
    dm = message.DataMessage(dataflow=df, data=[ds])

    obs = []
    for foo, bar in product(cl["FOO"], cl["BAR"]):
        # TODO Add attribute values
        obs.append(
            v21.Observation(
                attached_attribute={
                    "ATTR_BAZ": common.AttributeValue(
                        value=cl["BAZ"]["BAZ0"], value_for=ad.get("ATTR_BAZ")
                    ),
                    "ATTR_QUX": common.AttributeValue(
                        value=cl["QUX"]["QUX0"], value_for=ad.get("ATTR_QUX")
                    ),
                },
                dimension=common.Key(described_by=dd, DIM_FOO=foo, DIM_BAR=bar),
                value="1.23",
                value_for=md.components[0],
            )
        )
    ds.add_obs(obs)
    return sm, dm


#: Mapping from format_options=… argument → expected initial columns.
EXP_COLS_START = {
    csv.v1.FormatOptions: ["DATAFLOW"],
    csv.v2.FormatOptions: ["STRUCTURE", "STRUCTURE_ID", "ACTION"],
    type(None): [
        "STRUCTURE",
        "STRUCTURE_ID",
        "ACTION",
    ],  # Default: same as SDMX-CSV 2.x
}

#: Mapping from labels= keyword argument → expected column names.
EXP_COLS = {
    Labels.id: [
        "DIM_FOO",
        "DIM_BAR",
        "OBS_VALUE_X",
        "ATTR_BAZ",
        "ATTR_QUX",
    ],
    Labels.both: [
        "DIM_FOO: Foo",
        "DIM_BAR: Bar",
        "OBS_VALUE_X: Observation value",
        "ATTR_BAZ: Baz",
        "ATTR_QUX: Qux",
    ],
    Labels.name: [
        "DIM_FOO",
        "Foo",
        "DIM_BAR",
        "Bar",
        "OBS_VALUE_X",
        "Observation value",
        "ATTR_BAZ",
        "Baz",
        "ATTR_QUX",
        "Qux",
    ],
}


@pytest.mark.parametrize(
    "keys",
    (
        Keys.none,
        pytest.param(Keys.both, marks=pytest.mark.xfail(raises=NotImplementedError)),
        pytest.param(Keys.obs, marks=pytest.mark.xfail(raises=NotImplementedError)),
        pytest.param(Keys.both, marks=pytest.mark.xfail(raises=NotImplementedError)),
    ),
)
def test_write_keys(
    messages: tuple[message.StructureMessage, message.DataMessage], keys: Keys
) -> None:
    """SDMX-CSV can be produced with :attr:`Labels.both` and :attr:`Labels.name`."""
    sm, dm = messages

    fo = csv.v2.FormatOptions()
    result = sdmx.to_csv(
        dm, rtype=pd.DataFrame, format_options=fo, keys=keys, attributes=Attributes.all
    )
    assert isinstance(result, pd.DataFrame)

    assert EXP_COLS_START[type(fo)] + EXP_COLS[Labels.id] == result.columns.to_list()
    assert len(dm.data[0]) == len(result)


@pytest.mark.parametrize("fo", [None, csv.v1.FormatOptions(), csv.v2.FormatOptions()])
@pytest.mark.parametrize("labels", list(Labels))
def test_write_labels(
    messages: tuple[message.StructureMessage, message.DataMessage],
    fo: CSVFormatOptions,
    labels: Labels,
) -> None:
    """SDMX-CSV can be produced with :attr:`Labels.both` and :attr:`Labels.name`."""
    sm, dm = messages

    if type(fo) is csv.v1.FormatOptions and labels is Labels.name:
        pytest.skip(reason="Invalid combination")

    result = sdmx.to_csv(
        dm,
        rtype=pd.DataFrame,
        format_options=fo,
        labels=labels,
        attributes=Attributes.all,
    )
    assert isinstance(result, pd.DataFrame)

    # print(result.to_string())  # DEBUG
    expr = re.compile("_X" if type(fo) is csv.v1.FormatOptions else "^$")
    assert (
        list(map(partial(expr.sub, ""), EXP_COLS_START[type(fo)] + EXP_COLS[labels]))
        == result.columns.to_list()
    )
    assert len(dm.data[0]) == len(result)


def test_rtype_str(tmp_path, specimen):
    with specimen("ECB_EXR/1/M.USD.EUR.SP00.A.xml") as f:
        msg = sdmx.read_sdmx(f)
    ds = msg.data[0]
    _add_test_dsd(ds)

    isinstance(sdmx.to_csv(ds, rtype=str), str)


def test_unsupported(tmp_path, specimen):
    with specimen("ECB_EXR/1/M.USD.EUR.SP00.A.xml") as f:
        msg = sdmx.read_sdmx(f)
    ds = msg.data[0]

    with pytest.raises(ValueError, match="No associated data flow definition for"):
        sdmx.to_csv(ds)

    _add_test_dsd(ds)

    with pytest.raises(ValueError, match="rtype"):
        sdmx.to_csv(ds, rtype=int)

    with pytest.raises(TypeError, match="positional"):
        sdmx.to_csv(ds, "foo")

    with pytest.raises(NotImplementedError, match="time_format"):
        sdmx.to_csv(ds, time_format="normalized")



# Configuration file for the Sphinx documentation builder.
#
# For a full list of options, see
# https://www.sphinx-doc.org/en/master/usage/configuration.html

# -- Project information ---------------------------------------------------------------

project = "sdmx"
copyright = "2014–%Y sdmx1 developers"
author = "sdmx1 developers"


# -- General configuration -------------------------------------------------------------

# Sphinx extension module names
extensions = [
    "sphinx.ext.autodoc",
    "sphinx.ext.autosummary",
    "sphinx.ext.extlinks",
    "sphinx.ext.intersphinx",
    "sphinx.ext.linkcode",
    "sphinx.ext.napoleon",
    "sphinx.ext.todo",
    "IPython.sphinxext.ipython_console_highlighting",
    "IPython.sphinxext.ipython_directive",
]

nitpicky = True

rst_prolog = """
.. role:: py(code)
   :language: python
.. role:: xml(code)
   :language: xml
"""

# -- Options for HTML output -----------------------------------------------------------

GH_URL = "https://github.com/khaeru/sdmx"

html_theme = "furo"

html_theme_options = dict(
    footer_icons=[
        {
            "name": "GitHub",
            "url": GH_URL,
            "html": """
                <svg stroke="currentColor" fill="currentColor" stroke-width="0" viewBox="0 0 16 16">
                    <path fill-rule="evenodd" d="M8 0C3.58 0 0 3.58 0 8c0 3.54 2.29 6.53 5.47 7.59.4.07.55-.17.55-.38 0-.19-.01-.82-.01-1.49-2.01.37-2.53-.49-2.69-.94-.09-.23-.48-.94-.82-1.13-.28-.15-.68-.52-.01-.53.63-.01 1.08.58 1.23.82.72 1.21 1.87.87 2.33.66.07-.52.28-.87.51-1.07-1.78-.2-3.64-.89-3.64-3.95 0-.87.31-1.59.82-2.15-.08-.2-.36-1.02.08-2.12 0 0 .67-.21 2.2.82.64-.18 1.32-.27 2-.27.68 0 1.36.09 2 .27 1.53-1.04 2.2-.82 2.2-.82.44 1.1.16 1.92.08 2.12.51.56.82 1.27.82 2.15 0 3.07-1.87 3.75-3.65 3.95.29.25.54.73.54 1.48 0 1.07-.01 1.93-.01 2.2 0 .21.15.46.55.38A8.013 8.013 0 0 0 16 8c0-4.42-3.58-8-8-8z"></path>
                </svg>
            """,
            "class": "",
        },
    ],
    top_of_page_buttons=["view", "edit"],
    source_repository=GH_URL,
    source_branch="main",
    source_directory="doc/",
)

html_title = "sdmx1"

# -- Options for sphinx.ext.extlinks ---------------------------------------------------

extlinks = {
    "issue": ("https://github.com/khaeru/sdmx/issues/%s", "#%s"),
    "pull": ("https://github.com/khaeru/sdmx/pull/%s", "PR #%s"),
    "gh-user": ("https://github.com/%s", "@%s"),
}


# -- Options for sphinx.ext.intersphinx ------------------------------------------------

intersphinx_mapping = {
    "np": ("https://numpy.org/doc/stable/", None),
    "packaging": ("https://packaging.pypa.io/en/stable", None),
    "pd": ("https://pandas.pydata.org/pandas-docs/stable/", None),
    "py": ("https://docs.python.org/3/", None),
    "requests": ("https://requests.readthedocs.io/en/latest/", None),
    "requests-cache": ("https://requests-cache.readthedocs.io/en/latest/", None),
}

# -- Options for sphinx.ext.linkcode ---------------------------------------------------

LINKCODE_ALIAS = {
    "sdmx/testing": "sdmx/testing/__init__",
}


def linkcode_resolve(domain, info):
    if domain != "py" or not info["module"]:
        return None
    filename = info["module"].replace(".", "/")
    filename = LINKCODE_ALIAS.get(filename, filename)
    return f"https://github.com/khaeru/sdmx/tree/main/{filename}.py"


# -- Options for sphinx.ext.napoleon ---------------------------------------------------

napoleon_preprocess_types = True

# -- Options for sphinx.ext.todo -------------------------------------------------------

# If True, todo and todolist produce output, else they produce nothing
todo_include_todos = True

# -- Options for IPython.sphinxext.ipython_directive -----------------------------------

# Specify if the embedded Sphinx shell should import Matplotlib and set the backend
ipython_mplbackend = ""

# -- Dynamic configuration


def setup(app):
    from importlib import import_module
    from pathlib import Path

    # Update files containing lists of classes
    for mod in "common", "v21", "v30":
        # prefix = f"{mod}." if mod != "common" else ""
        prefix = f"{mod}."
        path = Path(__file__).parent.joinpath("api", f"model-{mod}-list.rst")
        with open(path, "w") as f:
            f.write(".. This file is auto-generated by doc/conf.py.\n\n")
            f.writelines(
                f":obj:`~.{prefix}{name}`\n"
                for name in sorted(import_module(f"sdmx.model.{mod}").__all__)
            )



import io
import re
from pathlib import Path

import pytest

import sdmx
from sdmx.format import Version, xml
from sdmx.format.xml.common import _extracted_zipball
from sdmx.message import StructureMessage
from sdmx.model import v21


def test_ns_prefix():
    with pytest.raises(ValueError):
        xml.v21.ns_prefix("https://example.com")


def test_qname():
    assert f"{{{xml.v21.base_ns}/structure}}Code" == str(xml.v21.qname("str", "Code"))
    assert f"{{{xml.v30.base_ns}/structure}}Code" == str(xml.v30.qname("str", "Code"))


def test_tag_for_class():
    # ItemScheme is never written to XML; no corresponding tag name
    assert xml.v21.tag_for_class(v21.ItemScheme) is None


def test_class_for_tag():
    assert xml.v30.class_for_tag("str:DataStructure") is not None


@pytest.mark.network
@pytest.mark.parametrize("version", ["2.1", "3.0.0"])
def test_install_schemas(installed_schemas, version):
    """Test that XSD files are downloaded and ready for use in validation."""
    # Look for a couple of the expected files
    for schema_doc in ("SDMXCommon.xsd", "SDMXMessage.xsd"):
        assert installed_schemas.joinpath(version, schema_doc).exists()


@pytest.mark.network
def test_install_schemas_in_user_cache():
    """Test that XSD files are downloaded and ready for use in validation."""
    import platformdirs

    cache_dir = platformdirs.user_cache_path("sdmx") / "2.1"
    sdmx.install_schemas()

    # Look for a couple of the expected files
    files = ["SDMXCommon.xsd", "SDMXMessage.xsd"]
    for schema_doc in files:
        doc = cache_dir.joinpath(schema_doc)
        assert doc.exists(), (cache_dir, sorted(cache_dir.glob("*")))


@pytest.mark.parametrize("version", ["1", 1, None])
def test_install_schemas_invalid_version(version):
    """Ensure invalid versions throw ``NotImplementedError``."""
    with pytest.raises(NotImplementedError):
        sdmx.install_schemas(version=version)


@pytest.mark.flaky(reruns=5)
@pytest.mark.network
@pytest.mark.parametrize(
    "parts",
    [
        ("v21", "xml", "common", "common.xml"),
        ("v21", "xml", "demography", "demography.xml"),
        ("v21", "xml", "demography", "esms.xml"),
        ("ECB_EXR", "common.xml"),
        ("ECB_EXR", "ng-structure-full.xml"),
        ("ECB_EXR", "ng-structure.xml"),
        ("v21", "xml", "query", "query_cl_all.xml"),
        ("v21", "xml", "query", "response_cl_all.xml"),
        ("v21", "xml", "query", "query_esms_children.xml"),
        ("v21", "xml", "query", "response_esms_children.xml"),
    ],
)
def test_validate_xml_from_v2_1_samples(tmp_path, specimen, installed_schemas, parts):
    """Use official samples to ensure validation of v2.1 messages works correctly."""
    # Samples are somewhat spread out, and some are known broken so we pick a bunch
    with specimen(str(Path(*parts))) as sample:
        assert sdmx.validate_xml(sample, installed_schemas, version="2.1")


@pytest.fixture(scope="module")
def v30_zipball_path(installed_schemas):
    yield _extracted_zipball(Version["3.0.0"])


@pytest.mark.flaky(reruns=5)
@pytest.mark.network
@pytest.mark.parametrize(
    "parts",
    [
        # Samples are somewhat spread out, and some are known broken so we pick a bunch
        ("Codelist", "codelist.xml"),
        ("Codelist", "codelist - extended.xml"),
        ("Concept Scheme", "conceptscheme.xml"),
        ("Data Structure Definition", "ECB_EXR.xml"),
        ("Dataflow", "dataflow.xml"),
        ("Geospatial", "geospatial_geographiccodelist.xml"),
    ],
)
def test_validate_xml_from_v3_0_samples(installed_schemas, v30_zipball_path, parts):
    """Use official samples to ensure validation of v3.0 messages works correctly."""

    assert sdmx.validate_xml(
        v30_zipball_path.joinpath("samples", *parts), installed_schemas, version="3.0.0"
    )


@pytest.mark.flaky(reruns=5)
@pytest.mark.network
def test_validate_xml_invalid_doc(tmp_path, installed_schemas):
    """Ensure that an invalid document fails validation."""
    msg_path = tmp_path / "invalid.xml"

    # Generate a codelist to form a StructureMessage
    ECB = v21.Agency(id="ECB")

    cl = v21.Codelist(
        id="CL_COLLECTION",
        version="1.0",
        is_final=False,
        is_external_reference=False,
        maintainer=ECB,
        name={"en": "Collection indicator code list"},
    )

    # Add items
    CL_ITEMS = [
        dict(id="A", name={"en": "Average of observations through period"}),
        dict(id="B", name={"en": "Beginning of period"}),
        dict(id="B1", name={"en": "Child code of B"}),
    ]
    for info in CL_ITEMS:
        cl.items[info["id"]] = v21.Code(**info)

    msg = StructureMessage(codelist={cl.id: cl})

    msg_path.write_bytes(sdmx.to_xml(msg))

    assert sdmx.validate_xml(msg_path, schema_dir=installed_schemas.joinpath("2.1"))


def test_validate_xml_invalid_message_type(installed_schemas):
    """Ensure that an invalid document fails validation."""
    # Create a mangled structure message with its outmost tag changed to be invalid
    msg = StructureMessage()
    invalid_msg = io.BytesIO(
        re.sub(b"mes:Structure([ >])", rb"mes:FooBar\1", sdmx.to_xml(msg))
    )

    with pytest.raises(NotImplementedError, match="Validate non-SDMX root.*FooBar>"):
        sdmx.validate_xml(invalid_msg, installed_schemas)


@pytest.mark.parametrize("version", ["1", 1, None])
def test_validate_xml_invalid_version(version):
    """Ensure validation of invalid versions throw ``NotImplementedError``."""
    with pytest.raises(NotImplementedError):
        # This message doesn't exist, but the version should throw before it is used.
        sdmx.validate_xml("samples/common/common.xml", version=version)


def test_validate_xml_max_errors(caplog, installed_schemas):
    """Test :py:`validate_xml(..., max_errors=...)`."""
    msg = StructureMessage()
    invalid_msg = io.BytesIO(
        re.sub(b"<(mes:Structures)/>", rb"<\1><Foo/></\1><Bar/>", sdmx.to_xml(msg))
    )

    # Without max_errors, 2 messages are logged
    sdmx.validate_xml(invalid_msg, installed_schemas)
    assert 2 == len(caplog.messages)
    caplog.clear()

    # With the argument, only 1 message is logged
    sdmx.validate_xml(invalid_msg, installed_schemas, max_errors=1)
    assert 1 == len(caplog.messages)


def test_validate_xml_no_schemas(tmp_path, specimen):
    """Check that supplying an invalid schema path will raise ``ValueError``."""
    with specimen("IPI-2010-A21-structure.xml", opened=False) as msg_path:
        with pytest.raises(FileNotFoundError):
            sdmx.validate_xml(msg_path, schema_dir=tmp_path)



""":ref:`sdmx-csv` writer."""

from os import PathLike
from typing import TYPE_CHECKING, Optional, Union

import pandas as pd

from sdmx.convert.pandas import PandasConverter
from sdmx.format.csv import common, v2

if TYPE_CHECKING:
    from sdmx.message import DataMessage
    from sdmx.model.common import BaseDataSet


def to_csv(
    obj: Union["DataMessage", "BaseDataSet"],
    *,
    path: Optional[PathLike] = None,
    rtype: type[Union[str, pd.DataFrame]] = str,
    **kwargs,
) -> Union[None, str, pd.DataFrame]:
    """Convert an SDMX *obj* to SDMX-CSV.

    With :py:`rtype=pandas.DataFrame`, the returned object is **not necessarily** valid
    SDMX-CSV. In particular, writing this to file using :meth:`pandas.DataFrame.to_csv`
    will yield invalid SDMX-CSV, because pandas includes a CSV column corresponding to
    the index of the data frame. You must pass :py:`index=False` to disable this
    behaviour. With :py:`rtype=str` or when giving `path`, this is done automatically.

    Parameters
    ----------
    path : os.PathLike, optional
        Path to write an SDMX-CSV file.
    rtype :
        Return type; see below. Pass literally :class:`str` or
        :class:`pandas.DataFrame`; *not* an instance of either class.

    Other parameters
    ----------------
    kwargs :
        Keyword arguments passed to :class:`.PandasConverter` and
        :class:`.format.csv.v1.FormatOptions`.

    Returns
    -------
    None :
        if `path` is given.
    str :
        if `rtype` is :class:`str`.
    pd.DataFrame :
        if `rtype` is :class:`~pandas.DataFrame`.

    Raises
    ------
    NotImplementedError
        If `obj` is any class except :class:`.DataSet`; this is the only class for which
        the SDMX-CSV standard describes a format.

        The two optional parameters are exactly as described in the specification.

        Because SDMX-CSV includes a column with an identifier (partial URN) for the
        dataflow to which the data belong (the column is named differently according to
        format version; see :ref:`sdmx-csv`), it is mandatory that the
        :attr:`DataSet.described_by <.BaseDataSet.described_by>` attribute of `obj`
        gives an association to an object from which a :class:`.URN` can be constructed.

    NotImplementedError
        if the `kwargs` include :py:`labels="name"` or :py:`time_format="normalized"`.
    ValueError
        If :attr:`DataSet.described_by <.BaseDataSet.described_by>` is :any:`None`.

    See also
    --------
    :ref:`sdmx.writer.csv <writer-csv>`.
    """
    common.kwargs_to_format_options(kwargs, v2.FormatOptions)
    result = PandasConverter(**kwargs).convert(obj)

    if path:
        return result.to_csv(path, index=False)
    elif rtype is str:
        return result.to_string(index=False)
    elif rtype is pd.DataFrame:
        return result
    else:
        raise ValueError(f"Invalid rtype={rtype!r}")



"""Information about SDMX-CSV file formats."""

import operator
from dataclasses import dataclass, fields, replace
from enum import Enum, Flag, auto
from functools import reduce

from sdmx.format.common import Format, FormatOptions


class Attributes(Flag):
    """Attributes to include."""

    #: No attributes.
    none = 0

    #: Attributes attached to each :class:`Observation <.BaseObservation>`.
    observation = auto()
    o = observation

    #: Attributes attached to any (0 or 1) :class:`~.SeriesKey` associated with each
    #: Observation.
    series_key = auto()
    s = series_key

    #: Attributes attached to any (0 or more) :class:`~.GroupKey` associated with each
    #: Observation.
    group_key = auto()
    g = group_key

    #: Attributes attached to the :class:`~.DataSet` containing the Observations.
    dataset = auto()
    d = dataset

    all = observation | series_key | group_key | dataset

    @classmethod
    def parse(cls, value: str) -> "Attributes":
        try:
            values = [Attributes.none] + [cls[v] for v in value.lower()]
        except KeyError as e:
            raise ValueError(f"{e.args[0]!r} is not a member of {cls}")
        return reduce(operator.or_, values)


class Labels(Enum):
    """SDMX-CSV 'labels' parameter."""

    #: Display only :attr:`IdentifiableArtefact.id`, for :class:`.Dimension` or
    #: :class:`.DataAttribute` in column headers and :class:`.Code` in data rows.
    id = auto()

    #: Display both the ID and the localized :attr:`NameableArtefact.name`.
    both = auto()

    #: Display only the localized name. Not present in SDMX-CSV 1.0
    name = auto()  # type: ignore [assignment]


class TimeFormat(Enum):
    """SDMX-CSV 'timeFormat' parameter."""

    #: Values for any dimension or attribute with ID ``TIME_PERIOD`` are displayed
    #: as recorded.
    original = auto()

    #: ``TIME_PERIOD`` values are converted to the most granular ISO 8601 representation
    #: taking into account the highest frequency of the data in the message and the
    #: moment in time when the lower-frequency values were collected.
    normalized = auto()


@dataclass
class CSVFormat(Format):
    """Information about an SDMX-CSV format."""

    suffix = "csv"


@dataclass
class CSVFormatOptions(FormatOptions):
    """SDMX-CSV format options.

    These options and default values are common to SDMX-CSV 1.0, 2.0.0, and 2.1.0.
    """

    format = CSVFormat

    #: Types of labels included.
    labels: Labels = Labels.id

    #: Time format.
    time_format: TimeFormat = TimeFormat.original

    def __post_init__(self) -> None:
        # Convert string arguments to enum members
        if isinstance(self.labels, str):
            self.labels = Labels[str(self.labels).lower()]
        if isinstance(self.time_format, str):
            self.time_format = TimeFormat[str(self.time_format).lower()]


def kwargs_to_format_options(kwargs: dict, cls: type["CSVFormatOptions"]) -> None:
    """Separate from `kwargs` any attributes of :class:`CSVFormatOptions`."""
    _fo = "format_options"
    # Either an existing instance of CSVFormatOptions or a subclass, or a new/default
    # instance
    existing = kwargs.pop(_fo, cls()) or cls()

    # Keys in `kwargs` that should be assigned to a FormatOptions instance
    keys = set(f.name for f in fields(existing)) & set(kwargs)
    replacements = {k: kwargs.pop(k) for k in keys}

    # Replace or set existing instance with `existing` + `replacements`
    kwargs[_fo] = replace(existing, **replacements)



"""Code for working with :file:`sdmx-test-data`."""

import logging
from contextlib import contextmanager
from pathlib import Path, PurePosixPath
from typing import TYPE_CHECKING, Optional

import platformdirs
import pytest

if TYPE_CHECKING:
    from sdmx.session import Session
    from sdmx.source import Source

log = logging.getLogger(__name__)

#: Default directory for local copy of sdmx-test-data.
DEFAULT_DIR = platformdirs.user_cache_path("sdmx").joinpath("test-data")

#: Expected to_pandas() results for data files; see :meth:`SpecimenCollection.expected_data`
#:
#: - Keys are the file name (see :func:`.add_specimens`) with '.' -> '-': 'foo.xml' ->
#:   'foo-xml'.
#: - Data is stored in :file:`sdmx-test-data/expected/{KEY}.txt`.
#: - Values are either argument to :func:`pandas.read_csv`; or a dict(use='other-key'),
#:   in which case the info for other-key is used instead.
EXPECTED = {
    "ng-flat-xml": dict(index_col=[0, 1, 2, 3, 4, 5]),
    "ng-ts-gf-xml": dict(use="ng-flat-xml"),
    "ng-ts-xml": dict(use="ng-flat-xml"),
    "ng-xs-xml": dict(index_col=[0, 1, 2, 3, 4, 5]),
    # Excluded: this file contains two DataSets, and
    # :meth:`SpecimenCollection.expected_data` currently only supports specimens with
    # one DataSet.
    # 'action-delete-json': dict(header=[0, 1, 2, 3, 4]),
    "xs-json": dict(index_col=[0, 1, 2, 3, 4, 5]),
    "flat-json": dict(index_col=[0, 1, 2, 3, 4, 5]),
    "ts-json": dict(use="flat-json"),
}

ERROR = (
    "Unable to locate test specimens. Give --sdmx-fetch-data, or use --sdmx-test-data=…"
    "or the SDMX_TEST_DATA environment variable to indicate an existing directory"
)

#: Git remote URL for cloning test data.
REMOTE_URL = "https://github.com/khaeru/sdmx-test-data.git"
# REMOTE_URL = "git@github.com:khaeru/sdmx-test-data.git"


class SpecimenCollection:
    """Collection of test specimens.

    Parameters
    ----------
    path :
       Path containing :file:`sdmx-test-data`, or to which to clone the repository.
    fetch :
       If :any:`True`, call :meth:`fetch`
    """

    #: Base path containing the specimen collection.
    base_path: Path

    #: Each tuple contains:
    #:
    #: 1. Path to specimen file.
    #: 2. Format: one of "csv", "json", "xml".
    #: 3. Message type: either "data", "structure", or None.
    specimens: list[tuple[Path, str, Optional[str]]]

    def __init__(self, path: Path, fetch: bool):
        self.base_path = path

        if fetch:
            self.fetch()

        # Check the base directory exists
        if not self.base_path.exists():  # pragma: no cover
            # Cannot proceed further; this exception kills the test session
            raise FileNotFoundError(f"SDMX test data in {path}\n{ERROR}")

        self.specimens = []
        add_specimens(self.specimens, self.base_path)

    @contextmanager
    def __call__(self, pattern="", opened: bool = True):
        """Open the test specimen file with `pattern` in the name.

        With :py:`opened=True` (default), return a file-like object. With
        :py:`opened=False`, return only the full path to the specimen file.
        """
        for path, f, k in self.specimens:
            if path.match("*" + pattern + "*"):
                yield open(path, "br") if opened else path
                return
        raise ValueError(pattern)  # pragma: no cover

    def as_params(
        self,
        format: Optional[str] = None,
        kind: Optional[str] = None,
        marks: Optional[dict] = None,
    ):
        """Generate :func:`pytest.param` from specimens.

        One :func:`~.pytest.param` is generated for each specimen that matches the
        `format` and `kind` arguments (if any). Marks are attached to each param from
        `marks`, wherein the keys are partial paths.
        """
        # Transform `marks` into a platform-independent mapping from path parts
        _marks = {PurePosixPath(k).parts: v for k, v in (marks or {}).items()}

        for path, f, k in self.specimens:
            if (format and format != f) or (kind and kind != k):
                continue
            p_rel = path.relative_to(self.base_path)
            yield pytest.param(
                path,
                id=str(p_rel),  # String ID for this specimen
                marks=_marks.get(p_rel.parts, tuple()),  # Look up marks via path parts
            )

    def expected_data(self, path):
        """Return the expected :func:`.to_pandas` result for the specimen `path`.

        Data is retrieved from :data:`.EXPECTED`.
        """
        import pandas as pd

        try:
            key = path.name.replace(".", "-")
            info = EXPECTED[key]
            if "use" in info:
                # Use the same expected data as another file
                key = info["use"]
                info = EXPECTED[key]
        except KeyError:
            return None

        args = dict(sep=r"\s+", index_col=[0], header=[0])
        args.update(info)

        result = pd.read_csv(
            self.base_path.joinpath("expected", key).with_suffix(".txt"), **args
        )

        # A series; unwrap
        if set(result.columns) == {"value"}:
            result = result["value"]

        return result

    def fetch(self) -> None:
        """Fetch test data from GitHub."""
        import git

        # Create a lock to avoid concurrency issues when running with pytest-xdist
        self.base_path.mkdir(parents=True, exist_ok=True)
        blf = git.BlockingLockFile(self.base_path, check_interval_s=0.1)
        blf._obtain_lock()

        # Initialize a git Repo object
        repo = git.Repo.init(self.base_path)

        try:
            # Reference to existing 'origin' remote
            origin = repo.remotes["origin"]
            # Ensure the REMOTE_URL is among the URLs for this remote
            if REMOTE_URL not in origin.urls:  # pragma: no cover
                origin.set_url(REMOTE_URL)
        except IndexError:
            # Create a new remote
            origin = repo.create_remote("origin", REMOTE_URL)

        log.info(f"Fetch test data from {origin} → {repo.working_dir}")

        origin.fetch("refs/heads/main", depth=1)  # Fetch only 1 commit from the remote
        origin_main = origin.refs["main"]  # Reference to 'origin/main'
        try:
            head = repo.heads["main"]  # Reference to existing local 'main'
        except IndexError:
            head = repo.create_head("main", origin_main)  # Create a local 'main'

        if (
            head.commit != origin_main.commit  # Commit differs
            or repo.is_dirty()  # Working dir is dirty
            or len(repo.index.diff(head.commit))
        ):
            # Check out files into the working directory
            head.set_tracking_branch(origin_main).checkout()

        del blf  # Release lock

    def parametrize(self, metafunc) -> None:
        """Handle the ``parametrize_specimens`` mark for a specific test."""
        try:
            mark = next(metafunc.definition.iter_markers("parametrize_specimens"))
        except StopIteration:
            return

        metafunc.parametrize(mark.args[0], self.as_params(**mark.kwargs))


def add_responses(session: "Session", file_cache_path: Path, source: "Source") -> None:
    """Populate cached responses for `session`.

    Two sources are used:

    1. Responses stored in :file:`sdmx-test-data/responses/`, as indicated by
       `file_cache_path`.
    2. For the ``TEST`` source as indicated by `source`, responses generated by this
       function. These are not stored in sdmx-test-data.
    """

    from requests_cache import FileCache

    import sdmx
    from sdmx.format import MediaType
    from sdmx.message import StructureMessage
    from sdmx.util.requests import save_response

    # Access the file cache in the given directory
    fc = FileCache(file_cache_path)

    # Add to `session` cache
    session.cache.update(fc)
    session.cache.recreate_keys()

    content: bytes = sdmx.to_xml(StructureMessage())
    headers = {"Content-Type": repr(MediaType("generic", "xml", "2.1"))}

    for endpoint in (
        "actualconstraint",
        "agencyscheme",
        "allowedconstraint",
        "attachementconstraint",
        "availableconstraint",
        "categorisation",
        "categoryscheme",
        "codelist",
        "conceptscheme",
        "contentconstraint",
        "customtypescheme",
        "dataconsumerscheme",
        "dataflow",
        "dataproviderscheme",
        "datastructure",
        "hierarchicalcodelist",
        "metadataflow",
        "metadatastructure",
        "namepersonalisationscheme",
        "organisationscheme",
        "organisationunitscheme",
        "process",
        "provisionagreement",
        "reportingtaxonomy",
        "rulesetscheme",
        "schema/datastructure",
        "structure",
        "structureset",
        "transformationscheme",
        "userdefinedoperatorscheme",
        "vtlmappingscheme",
    ):
        url = f"{source.url}/{endpoint}/{source.id}/all/latest"
        save_response(session, method="GET", url=url, content=content, headers=headers)

    for url in (
        f"{source.url}/availableconstraint",
        f"{source.url}/categoryscheme/{source.id}/all/latest?references=parentsandsiblings",
    ):
        save_response(session, method="GET", url=url, content=content, headers=headers)


def add_specimens(target: list[tuple[Path, str, Optional[str]]], base: Path) -> None:
    """Populate the `target` collection with specimens from :file:`sdmx-test-data`."""
    # XML data files for the ECB exchange rate data flow
    for source_id in ("ECB_EXR",):
        for path in base.joinpath(source_id).rglob("*.xml"):
            kind = "data"
            if "structure" in path.name or "common" in path.name:
                kind = "structure"
            target.append((path, "xml", kind))

    # JSON data files for ECB and OECD data flows
    for source_id in ("ECB_EXR", "OECD"):
        target.extend(
            (fp, "json", "data") for fp in base.joinpath(source_id).rglob("*.json")
        )

    # Miscellaneous XML data files
    target.extend(
        (base.joinpath(*parts), "xml", "data")
        for parts in [
            ("constructed", "gh-218.xml"),
            ("INSEE", "CNA-2010-CONSO-SI-A17.xml"),
            ("INSEE", "IPI-2010-A21.xml"),
            ("IMF", "PCPS.xml"),
            ("ESTAT", "demography-xs.xml"),
            ("ESTAT", "esms.xml"),
            ("ESTAT", "footer.xml"),
            ("ESTAT", "NAMA_10_GDP-ss.xml"),
        ]
    )

    # Miscellaneous XML structure files
    target.extend(
        (base.joinpath(*parts), "xml", "structure")
        for parts in [
            ("BIS", "actualconstraint-0.xml"),
            ("BIS", "hierarchicalcodelist-0.xml"),
            ("BIS", "gh-180.xml"),
            ("ECB", "orgscheme.xml"),
            ("ECB", "structureset-0.xml"),
            ("ESTAT", "apro_mk_cola-structure.xml"),
            ("ESTAT", "demography-structure.xml"),
            ("ESTAT", "esms-structure.xml"),
            ("ESTAT", "GOV_10Q_GGNFA.xml"),
            ("ESTAT", "HCL_WSTATUS_SCL_BNSPART.xml"),
            ("ESTAT", "HCL_WSTATUS_SCL_WSTATUSPR.xml"),
            ("IAEG-SDGs", "metadatastructure-0.xml"),
            ("IMF", "01R.xml"),
            ("IMF", "1PI-structure.xml"),
            ("IMF", "CL_AREA-structure.xml"),
            ("IMF", "CL_FREQ-3.0-structure.xml"),
            ("IMF", "datastructure-0.xml"),
            # Manually reduced subset of the response for this DSD. Test for
            # <str:CubeRegion> containing both <com:KeyValue> and <com:Attribute>
            ("IMF", "ECOFIN_DSD-structure.xml"),
            ("IMF", "hierarchicalcodelist-0.xml"),
            ("IMF", "hierarchicalcodelist-1.xml"),
            ("IMF", "structureset-0.xml"),
            ("IMF_STA", "availableconstraint_CPI.xml"),  # khaeru/sdmx#161
            ("IMF_STA", "DSD_GFS.xml"),  # khaeru/sdmx#164
            ("INSEE", "CNA-2010-CONSO-SI-A17-structure.xml"),
            ("INSEE", "dataflow.xml"),
            ("INSEE", "gh-205.xml"),
            ("INSEE", "IPI-2010-A21-structure.xml"),
            ("ISTAT", "22_289-structure.xml"),
            ("ISTAT", "47_850-structure.xml"),
            ("ISTAT", "actualconstraint-0.xml"),
            ("ISTAT", "metadataflow-0.xml"),
            ("ISTAT", "metadatastructure-0.xml"),
            ("OECD", "actualconstraint-0.xml"),
            ("OECD", "metadatastructure-0.xml"),
            ("UNICEF", "GLOBAL_DATAFLOW-structure.xml"),
            ("UNSD", "codelist_partial.xml"),
            ("SDMX", "HCL_TEST_AREA.xml"),
            ("SGR", "common-structure.xml"),
            ("SGR", "hierarchicalcodelist-0.xml"),
            ("SGR", "metadatastructure-0.xml"),
            ("SPC", "actualconstraint-0.xml"),
            ("SPC", "metadatastructure-0.xml"),
            ("TEST", "gh-142.xml"),
            ("TEST", "gh-149.xml"),
            ("WB", "gh-78.xml"),
        ]
    )

    # Files from the SDMX 2.1 specification
    v21 = base.joinpath("v21", "xml")
    target.extend((p, "xml", None) for p in v21.glob("**/*.xml"))

    # Files from the SDMX 3.0 specification
    v3 = base.joinpath("v3")

    # Files from the SDMX-CSV 2.0.0 specification
    target.extend((p, "csv", "data") for p in v3.joinpath("csv").glob("*.csv"))

    # commented: SDMX-JSON 2.0 is not yet implemented
    # # SDMX-JSON
    # self.specimens.extend(
    #     (p, "json", "data") for p in v3.joinpath("json", "data").glob("*.json")
    # )
    # for dir in ("metadata", "structure"):
    #     self.specimens.extend(
    #         (p, "json", "structure")
    #         for p in v3.joinpath("json", dir).glob("*.json")
    #     )

    # SDMX-ML
    target.extend((p, "xml", None) for p in v3.glob("xml/*.xml"))



from io import BufferedIOBase, BufferedRandom, BytesIO
from typing import IO, TYPE_CHECKING, Union

import requests

from sdmx.util.requests import HAS_REQUESTS_CACHE, CacheMixin, SessionAttrs

if TYPE_CHECKING:
    import os


class Session(CacheMixin, requests.Session):
    """:class:`requests.Session` with optional caching.

    If :mod:`requests_cache` is installed, this class inherits from
    :class:`~.requests_cache.CacheMixin` and caches responses. Otherwise, it inherits
    only from the :class:`requests_cache`.

    Parameters
    ----------
    timeout : float
        Timeout in seconds, used for every request.

    Other parameters
    ----------------
    kwargs :
        These may include:

        1. Values for any attributes of :class:`requests.Session`, such as
           :attr:`~requests.Session.proxies`,
           :attr:`~requests.Session.stream`, or
           :attr:`~requests.Session.verify`. These are set on the created Session.
        2. Keyword arguments to :class:`~.requests_cache.CacheMixin` or any
           :mod:`requests_cache` backend. Note that:

           - Unlike :mod:`requests_cache`, you must supply :py:`backend="sqlite"`
             explicitly; otherwise :mod:`sdmx` uses :py:`backend="memory"`.
           - These classes will silently ignore any other/unrecognized keyword
             arguments.

    Raises
    ------
    TypeError
        if :mod:`requests_cache` is *not* installed and any parameters are passed except
        for `timeout`.
    """

    timeout: float

    def __init__(self, timeout: float = 30.0, **kwargs):
        # Store timeout; not an attribute of requests.Session
        self.timeout = timeout

        # Separate kwargs that will update base requests.Session attributes
        attrs = [
            (k, kwargs.pop(k)) for k in set(SessionAttrs.__annotations__) & set(kwargs)
        ]

        if HAS_REQUESTS_CACHE:
            # Disable caching entirely if no cache-related kwargs are supplied
            kwargs.setdefault("disabled", not len(kwargs))
            kwargs.setdefault("backend", "memory")  # Avoid creating any file
        elif len(kwargs):  # pragma: no cover
            raise TypeError(
                f"Keyword arguments for not installed requests_cache: {kwargs}"
            )

        super().__init__(**kwargs)

        # Update attributes set by requests.Session.__init__()
        for name, value in attrs:
            setattr(self, name, value)


class ResponseIO(BufferedIOBase):
    """Buffered wrapper for :class:`requests.Response` with optional file output.

    :class:`ResponseIO` wraps a :class:`requests.Response` object's 'content'
    attribute, providing a file-like object from which bytes can be :meth:`read`
    incrementally.

    Parameters
    ----------
    response : :class:`requests.Response`
        HTTP response to wrap.
    tee : binary, writable :py:class:`io.BufferedIOBase`, defaults to io.BytesIO()
        *tee* is exposed as *self.tee* and not closed explicitly.
    """

    tee: IO

    def __init__(self, response, tee: Union[IO, "os.PathLike", None] = None):
        self.response = response

        if tee is None:
            self.tee = BytesIO()
        elif isinstance(tee, (IO, BufferedRandom)):
            # If tee is a file-like object or tempfile, then use it as cache
            self.tee = tee
        else:
            # So tee must be str, pathlib.Path, or similar
            self.tee = open(tee, "w+b")

        content_disposition = response.headers.get("Content-Disposition", "")
        if content_disposition.endswith('.gz"'):
            import gzip

            content = gzip.GzipFile(fileobj=BytesIO(response.content)).read()
        else:
            content = response.content

        self.tee.write(content)
        self.tee.seek(0)

    def readable(self):
        return True

    def read(self, size=-1):
        """Read and return up to `size` bytes by calling ``self.tee.read()``."""
        return self.tee.read(size)



from . import Source as BaseSource


class Source(BaseSource):
    """Hooks for the ``IMF_DATA`` SDMX-REST API."""

    _id = "IMF_DATA"

    def modify_request_args(self, kwargs):
        """Modify arguments used to build query URL.

        Set default provider agency ID ``all``.
        """
        super().modify_request_args(kwargs)

        # Supply 'all' as the default agency_id
        # NB this is an indirect test for resource_type != 'data'; because of the way
        #    the hook is called, resource_type is not available directly.
        if "key" not in kwargs:
            kwargs.setdefault("agency_id", "all")



from requests.structures import CaseInsensitiveDict

from sdmx.format import MediaType

from . import Source as BaseSource


class Source(BaseSource):
    """Hooks for the ``IMF_DATA3`` SDMX-REST API."""

    _id = "IMF_DATA3"

    def modify_request_args(self, kwargs):
        """Modify arguments used to build query URL.

        1. Set default provider agency ID ``all``.
        2. Set media-type in header.
        """
        super().modify_request_args(kwargs)

        # Supply 'all' as the default agency_id
        # NB this is an indirect test for resource_type != 'data'; because of the way
        #    the hook is called, resource_type is not available directly.
        if "key" not in kwargs:
            kwargs.setdefault("agency_id", "all")

        # Retrieve SDMX-ML by default
        # TODO Choose between the data, metadata, or structure media-type according to
        #      other `kwargs`
        kwargs.setdefault("headers", CaseInsensitiveDict())
        mt = [MediaType(x, "xml", "3.0.0") for x in ("data", "metadata", "structure")]
        kwargs["headers"].setdefault("Accept", ", ".join(map(str, mt)))



import logging
import os
import re
from collections import ChainMap
from collections.abc import Generator, Iterator
from copy import deepcopy
from pathlib import Path
from typing import TYPE_CHECKING, Union

import numpy as np
import pandas as pd
import pytest
import responses
from filelock import FileLock
from xdist import is_xdist_worker

from sdmx.exceptions import HTTPError
from sdmx.format import Version
from sdmx.rest import Resource
from sdmx.session import Session
from sdmx.source import DataContentType, Source, get_source
from sdmx.testing.report import ServiceReporter
from sdmx.util.requests import offline

from . import data

if TYPE_CHECKING:
    import pytest

log = logging.getLogger(__name__)

# Pytest stash keys
KEY_DATA = pytest.StashKey[Path]()
KEY_SPECIMENS = pytest.StashKey["data.SpecimenCollection"]()
KEY_SOURCE = pytest.StashKey["Source"]()


def assert_pd_equal(left, right, **kwargs):
    """Assert equality of two pandas objects."""
    if left is None:
        return
    method = {
        pd.Series: pd.testing.assert_series_equal,
        pd.DataFrame: pd.testing.assert_frame_equal,
        np.ndarray: np.testing.assert_array_equal,
    }[left.__class__]
    method(left, right, **kwargs)


def pytest_addoption(parser):
    """Add pytest command-line options."""
    parser.addoption(
        "--sdmx-fetch-data",
        action="store_true",
        help="fetch test specimens from GitHub",
    )
    parser.addoption(
        "--sdmx-test-data",
        # Use the environment variable value by default
        default=os.environ.get("SDMX_TEST_DATA", data.DEFAULT_DIR),
        help="path to SDMX test specimens",
    )


def pytest_configure(config):
    """Handle the ``--sdmx-test-data`` command-line option."""
    # Register "parametrize_specimens" as a known mark to suppress warnings from pytest
    config.addinivalue_line(
        "markers", "parametrize_specimens: (for internal use by sdmx.testing)"
    )

    # Register plugin for reporting service outputs
    config._sdmx_reporter = ServiceReporter(config)
    config.pluginmanager.register(config._sdmx_reporter)

    # Convert the option value to an Path instance attribute on `config`
    try:
        config.stash[KEY_DATA] = Path(config.option.sdmx_test_data)
    except TypeError:  # pragma: no cover
        raise RuntimeError(data.ERROR) from None


def pytest_sessionstart(session: "pytest.Session") -> None:
    """Create session-wide objects.

    These are used by the fixtures :func:`.specimen`, :func:`.testsource`.
    """
    c = session.config

    # Create a SpecimenCollection from the files in the directory
    c.stash[KEY_SPECIMENS] = data.SpecimenCollection(
        c.stash[KEY_DATA],
        # Fetch the test data if so configured, and not in an xdist worker process
        c.option.sdmx_fetch_data and not is_xdist_worker(session),
    )

    # Create a test source
    c.stash[KEY_SOURCE] = Source(
        id="TEST",
        name="Test source",
        url="https://example.com/sdmx-rest",
        supports={feature: True for feature in list(Resource)},
    )


def pytest_generate_tests(metafunc):
    """Generate tests.

    Calls both :func:`parametrize_specimens` and :func:`generate_endpoint_tests`.
    """
    # Handle the 'parametrize_specimens' mark
    metafunc.config.stash[KEY_SPECIMENS].parametrize(metafunc)
    # Handle the 'endpoint' fixture request
    generate_endpoint_tests(metafunc)


#: Marks for use below.
XFAIL = {
    # Exceptions resulting from querying an endpoint not supported by a service
    "unsupported": pytest.mark.xfail(
        strict=True,
        reason="Not implemented by service",
        raises=(
            HTTPError,  # 401, 404, 405, etc.
            NotImplementedError,  # 501, converted automatically
            ValueError,  # e.g. WB_WDI, returns invalid content type
        ),
    ),
    # Returned by servers that may be temporarily unavailable at the time of test
    503: pytest.mark.xfail(
        raises=HTTPError, reason="503 Server Error: Service Unavailable"
    ),
}


def generate_endpoint_tests(metafunc):  # noqa: C901  TODO reduce complexity 11 → ≤10
    """pytest hook for parametrizing tests that need an "endpoint" fixture.

    This function relies on the :class:`.DataSourceTest` base class defined in
    :mod:`.test_sources`. It:

    - Generates one parametrization for every :class:`.Resource` (= REST API endpoint).
    - Applies pytest "xfail" (expected failure) marks according to:

      1. :attr:`.Source.supports`, i.e. if the particular source is marked as not
         supporting certain endpoints, the test is expected to fail.
      2. :attr:`.DataSourceTest.xfail`, any other failures defined on the source test
         class (e.g. :class:`.DataSourceTest` subclass).
      3. :attr:`.DataSourceTest.xfail_common`, common failures.
    """
    if "endpoint" not in metafunc.fixturenames:
        return  # Don't need to parametrize this metafunc

    # Arguments to parametrize()
    params = []

    # Use the test class' source_id attr to look up the Source class
    cls = metafunc.cls
    source = (
        get_source(cls.source_id)
        if cls.source_id != "TEST"
        else metafunc.config.stash[KEY_SOURCE]
    )

    # Merge subclass-specific and "common" xfail marks, preferring the former
    xfails = ChainMap(cls.xfail, cls.xfail_common)

    # Iterate over all known endpoints
    for ep in Resource:
        # Accumulate multiple marks; first takes precedence
        marks = []

        # Get any keyword arguments for this endpoint
        args = cls.endpoint_args.get(ep.name, dict())
        if ep is Resource.data and not len(args):
            # args must be specified for a data query; no args → no test
            continue

        # Check if the associated source supports the endpoint
        supported = source.supports[ep]
        if source.data_content_type == DataContentType.JSON and ep is not Resource.data:
            # SDMX-JSON sources only support data queries
            continue
        elif not supported:
            args["force"] = True
            marks.append(XFAIL["unsupported"])

        # Check if the test function's class contains an expected failure for `endpoint`
        xfail = xfails.get(ep.name, None)
        if not marks and xfail:
            # Mark the test as expected to fail
            try:  # Unpack a tuple
                mark = pytest.mark.xfail(raises=xfail[0], reason=xfail[1])
            except TypeError:
                mark = pytest.mark.xfail(raises=xfail)
            marks.append(mark)

            if not supported:  # pragma: no cover; for identifying extraneous entries
                log.info(
                    f"tests for {source.id!r} mention unsupported endpoint {ep.name!r}"
                )

        # Tolerate 503 errors
        if cls.tolerate_503:
            marks.append(XFAIL[503])

        params.append(pytest.param(ep, args, id=ep.name, marks=marks))

    if len(params):
        # Run the test function once for each endpoint
        metafunc.parametrize("endpoint, args", params)
    # commented: for debugging
    # else:
    #     pytest.skip("No endpoints to be tested")


class CompareTests:
    """Base class for testing of :meth:`.Comparable.compare` in subclasses.

    For usage, see :class:`.test_common.TestIdentifiableArtefact`.
    """

    def test_compare(self, left, callback) -> None:
        """Test comparison of `left` to a copy modified using `callback`."""
        # Make a copy
        right = deepcopy(left)

        if callback is None:
            expected = True  # No callback → should compare equal
        else:
            callback(right)  # Apply some modification to the copy
            expected = False  # Should compare different

        try:
            assert expected is left.compare(right)
        except Exception:  # pragma: no cover
            # Show information about the objects
            log.error(f"{left.__dict__ = !r}")
            log.error(f"{right.__dict__ = !r}")
            raise


class MessageTest:
    """Base class for tests of specific specimen files."""

    directory: Union[str, Path] = Path(".")
    filename: str

    @pytest.fixture(scope="class")
    def path(self, test_data_path):
        yield test_data_path / self.directory

    @pytest.fixture(scope="class")
    def msg(self, path):
        import sdmx

        return sdmx.read_sdmx(path / self.filename)


@pytest.fixture(scope="session")
def installed_schemas(worker_id, mock_gh_api, tmp_path_factory):
    """Fixture that ensures schemas are installed locally in a temporary directory."""
    from sdmx.format.xml.common import install_schemas

    # Determine a consistent path on each worker for schema installation
    if worker_id == "master":  # xdist controller *or* not using test distribution
        dir = tmp_path_factory.mktemp("schemas")  # pragma: no cover
    else:  # xdist worker: find relative to the parent of the basetemp for this worker
        dir = tmp_path_factory.getbasetemp().parent.joinpath("schemas")

    # Don't try to unpack from multiple workers at once
    with mock_gh_api, FileLock(dir.with_suffix(".lock")):
        install_schemas(dir.joinpath("2.1"), Version["2.1"])
        install_schemas(dir.joinpath("3.0.0"), Version["3.0.0"])

    yield dir


@pytest.fixture(scope="session")
def mock_gh_api():
    """Mock GitHub API responses to avoid hitting rate limits.

    For each API endpoint URL queried by :func:.`_gh_zipball`, return a pared-down JSON
    response that contains the required "zipball_url" key.
    """
    base = "https://api.github.com/repos/sdmx-twg/sdmx-ml"

    # TODO Improve .util.requests to provide (roughly) the same functionality, then drop
    # use of responses here
    mock = responses.RequestsMock(assert_all_requests_are_fired=False)
    mock.add_passthru(re.compile(rf"{base}/zipball/\w+"))
    mock.add_passthru(re.compile(r"https://codeload.github.com/\w+"))
    mock.add_passthru(re.compile(r"http://www.w3.org/\w+"))

    for v in "2.1", "3.0", "3.0.0":
        mock.get(
            url=f"{base}/releases/tags/v{v}",
            json=dict(zipball_url=f"{base}/zipball/v{v}"),
        )

    yield mock


@pytest.fixture(scope="session")
def session_with_pytest_cache(pytestconfig):
    """Fixture:  A :class:`.Session` that caches within :file:`.pytest_cache`.

    This subdirectory is ephemeral, and tests **must** pass whether or not it exists and
    is populated.
    """
    p = pytestconfig.cache.mkdir("sdmx-requests-cache")
    yield Session(cache_name=str(p), backend="filesystem")


@pytest.fixture(scope="session")
def session_with_stored_responses(pytestconfig):
    """Fixture: A :class:`.Session` returns only stored responses from sdmx-test-data.

    This session…

    1. uses the 'memory' :mod:`requests_cache` backend;
    2. contains the responses from :func:`.testing.data.add_responses`; and
    3. is treated with :func:`.offline`, so that *only* stored responses can be
       returned.
    """
    session = Session(backend="memory")

    data.add_responses(
        session,
        pytestconfig.stash[KEY_DATA].joinpath("recorded"),
        pytestconfig.stash[KEY_SOURCE],
    )

    # Raise an exception on any actual attempts to access the network
    offline(session)

    yield session


@pytest.fixture(scope="session")
def specimen(pytestconfig) -> Iterator["data.SpecimenCollection"]:
    """Fixture: the :class:`SpecimenCollection`."""
    yield pytestconfig.stash[KEY_SPECIMENS]


@pytest.fixture(scope="session")
def test_data_path(pytestconfig):
    """Fixture: the :py:class:`.Path` given as --sdmx-test-data."""
    yield pytestconfig.stash[KEY_DATA]


@pytest.fixture(scope="class")
def testsource(pytestconfig) -> Generator[str, None, None]:
    """Fixture: the :attr:`.Source.id` of a temporary data source."""
    from sdmx.source import sources

    s = pytestconfig.stash[KEY_SOURCE]

    sources[s.id] = s

    try:
        yield s.id
    finally:
        sources.pop(s.id)



"""SDMX-CSV 1.0 format."""

from dataclasses import dataclass

from sdmx.format.csv.common import CSVFormat, CSVFormatOptions, Labels


class FORMAT(CSVFormat):
    version = "1.0"


@dataclass
class FormatOptions(CSVFormatOptions):
    """Format options for SDMX-CSV version 1.0."""

    format = FORMAT

    def __post_init__(self) -> None:
        if self.labels is Labels.name:
            raise ValueError("Labels.name is not valid for SDMX-CSV 1.0")
